// converts direction tables from selfav to bond tables of
// connlist structs chopped to the right length in the out file.
// Input format limits nodes to 33. Can handle 4-33 w/o recompile.

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cstring>

const int NNODMAX = 34;
const int STARTP = NNODMAX, ARRW = NNODMAX*2+1;
short nnod, maxconns;
unsigned char connfromto[NNODMAX][NNODMAX];

struct connlist
{
  char isdeg;
  char conns[NNODMAX];
};

void initmaxconns(void)
{
   int sqrt4nnod;
   for(sqrt4nnod = 1; sqrt4nnod*sqrt4nnod < 4*nnod; sqrt4nnod++);
   maxconns = nnod - sqrt4nnod + 1;
}

void initconnfromto(void)
{
  int from, to;
  int connnr = 1;
	for(from = 0; from < nnod; from++)	{
		for(to = from + 3; to < nnod; to += 2) {
      connfromto[from][to] = connnr;
      connfromto[to][from] = connnr;
      connnr++;
    }
  }
  std::cout << connnr-1 << " different connections\n";
}

short movepos(int &x, int &y, int d)
{
   switch (d) {
     case 0: x++; return 1;
     case 1: x--; return 1;
     case 2: y++; return 1;
     case 3: y--; return 1;
   }
   return 0;
}

short readseq(short seq[], int& deg, std::ifstream &inf)
{
  static unsigned short pkseq[4];
  int i;

  inf.read((char *)pkseq,8);
  if(inf.eof()) return 0;

  for(i=0; i<nnod-2; i++)
    seq[i] = ( pkseq[i>>3] >> ((i & 7) << 1) ) & 3;

  deg = pkseq[3] >> 15;

  return 1;
}

int writeconnlist(short seq[], int isdeg, std::ofstream &outf)
{
  static int nrarr[ARRW][ARRW];
  static int ofs = 1;
  static connlist connl;
  int x=STARTP, y=STARTP, connnr=0, nr, nrto, nconns;
 
  ofs += nnod;
  if(ofs < 0) {std::cout << "offset negative in bseq\n"; exit(0);} 

  nrarr[x][y] = ofs ;
  x++;
  nrarr[x][y] = ofs - 1;

  for(nr=2; nr<nnod; nr++) {
    if(!movepos(x,y,seq[nr-2])){ std::cout << "Bug in 1 writeconnlist\n"; exit(0); }
    nrarr[x][y] = ofs - nr;
    if((nrto = ofs-nrarr[x+1][y]) < nr - 1) connl.conns[connnr++] = connfromto[nr][nrto];
    if((nrto = ofs-nrarr[x-1][y]) < nr - 1) connl.conns[connnr++] = connfromto[nr][nrto];
    if((nrto = ofs-nrarr[x][y+1]) < nr - 1) connl.conns[connnr++] = connfromto[nr][nrto];
    if((nrto = ofs-nrarr[x][y-1]) < nr - 1) connl.conns[connnr++] = connfromto[nr][nrto];
    if(connnr && !connl.conns[connnr-1]){ std::cout << "Bug in 2 writeconnlist\n"; exit(0); }
  }
  nconns = connnr;
  if(nconns > maxconns){ std::cout << "Bug in 3 writeconnlist\n"; exit(0); }

  for(; connnr < maxconns; connnr++)
    connl.conns[connnr] = 0;

  connl.isdeg = isdeg;

  outf.write((char *)&connl,maxconns+1);

  return nconns;
}

int main(int argc, char **argv)
{
   int i, nb, nbmax=0;
   char infstr[16], outfstr[16];
   short seq[NNODMAX];
   int fnbnd[NNODMAX], fnbndnotdeg[NNODMAX];
   int nbtot=0;
   int isdeg=0;

	if(argc != 2)
	{
		std::cerr << "Usage: "<<argv[0]<<" <N>\n";
		exit(1);
	}
	nnod = atoi(argv[1]);

   initmaxconns();
   std::cout << "Maximally " << maxconns << " connections\n";
   initconnfromto();


   for(i=0; i<nnod; i++)
     fnbnd[i]=fnbndnotdeg[i]=0;

   strcpy(infstr, "sa9n00.out");
   infstr[4] = '0' + nnod / 10;
   infstr[5] = '0' + nnod % 10;
   std::ifstream inf(infstr, std::ios::in|std::ios::binary);
   if(!inf){std::cerr << "Unable to open " << infstr << '\n'; exit(0);}

   strcpy(outfstr, "sab5n00.out");
   outfstr[5] = '0' + nnod / 10;
   outfstr[6] = '0' + nnod % 10;
   std::ofstream outf(outfstr, std::ios::out|std::ios::trunc);
   if(!outf){std::cerr << "Unable to create " << outfstr << '\n'; exit(0);}

   std::cout << infstr << " -> " << outfstr << "\n";

   while(readseq(seq, isdeg, inf)) {
     nbtot += nb = writeconnlist(seq, isdeg, outf);
     fnbnd[nb]++;
       fnbndnotdeg[nb] += 1-isdeg;
     if(nb>nbmax) nbmax = nb;
   }
   inf.close();
   outf.close();

   std::cout << "Frequency of connection sites \\ for non-degenererate\n";
   for(i=0; i<=nbmax; i++){
     std::cout << i << ": " << fnbnd[i] << " \\ " << fnbndnotdeg[i] << "\n";
   }
   std::cout << "Total # of connection sites: " << nbtot << '\n';
}

