// Generates direction table for chosen SAWs. Detected degeneracy is flagged
// in the output file. Output format is limited to 33 or fewer nodes.

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cstring>
#include <stdint.h>

const int ARRW = 70;
const int START = ARRW / 2;
int nnod;
short arr[ARRW][ARRW], nrarr[ARRW][ARRW];
int fnbond[ARRW], fnbpont[ARRW];
int nbmax, nbpmax, ldnposx, ldnposy;
int upp;
long long ntot, nnotdeg;

int nneib(int x, int y)
{
  int nn = 0;
  if(arr[x+1][y]) nn++;
  if(arr[x-1][y]) nn++;
  if(arr[x][y+1]) nn++;
  if(arr[x][y-1]) nn++;
  return nn;
}

int movepos(int &x, int &y)
{
   switch (arr[x][y]) {
     case 1: x++; return 1;
     case 2: x--; return 1;
     case 3: y++; return 1;
     case 4: y--; return 1;
   }
   return 0;
}

int dfrees()
{
  int x,y,nn=0;
  x=y=START;
  nn += 4-nneib(x,y);
  x++;
  nn += 4-nneib(x,y);
  if(nn > 5){std::cout << "Bug 1 in dfrees\n"; exit(0);}
  return nn == 5;
}

int dfreee()
{
  int x,y,nn=0;
  x=ldnposx;
  y=ldnposy;
  nn += 4-nneib(x,y);
  movepos(x,y);
  nn += 4-nneib(x,y);
  if(nn > 5){std::cout << "Bug 1 in dfreee\n"; exit(0);}
  return nn == 5;
}

int lpnr(int tnr, int nr)
{
  return tnr > 0 && tnr < nr;
}

int sfpossmin(int& deg)
{
  int x, y;
  int nbrb = 0;
  int nd13, nd23, nd24, nd14;
  int nb13, nb23, nb24, nb14;
  nd13 = nd23 = nd24 = nd14 = 0;
  nb13 = nb23 = nb24 = nb14 = 0;
  x = y = START;
  deg = 0;

  for(int nr = 1; nr <= nnod; nr++, movepos(x,y)) {
    if(nrarr[x+1][y] > nr+1) nbrb++;
    if(nrarr[x-1][y] > nr+1) nbrb++;
    if(nrarr[x][y+1] > nr+1) nbrb++;
    if(nrarr[x][y-1] > nr+1) nbrb++;
    if(lpnr(nrarr[x+1][y], nr-1)) nbrb--;
    if(lpnr(nrarr[x-1][y], nr-1)) nbrb--;
    if(lpnr(nrarr[x][y+1], nr-1)) nbrb--;
    if(lpnr(nrarr[x][y-1], nr-1)) nbrb--;
    if(nbrb < 0) {std::cout << "Bug 1 in sfpossmin\n"; exit(0); }
	
    //will be on later
    if(nrarr[x+1][y+1] > nr) nd13++;
    if(nrarr[x-1][y+1] > nr) nd23++;
    if(nrarr[x-1][y-1] > nr) nd24++;
    if(nrarr[x+1][y-1] > nr) nd14++;
    //were on earlier
    if(lpnr(nrarr[x-1][y-1], nr)) nd13--;
    if(lpnr(nrarr[x+1][y-1], nr)) nd23--;
    if(lpnr(nrarr[x+1][y+1], nr)) nd24--;
    if(lpnr(nrarr[x-1][y+1], nr)) nd14--;
    if(nd13<0 || nd23<0 || nd24<0 || nd14<0) {std::cout << "Bug 2 in sfpossmin\n"; exit(0);}

    //binder till senare
    if(nrarr[x+1][y] > nr) {nb13++; nb14++;}
    if(nrarr[x-1][y] > nr) {nb23++; nb24++;}
    if(nrarr[x][y+1] > nr) {nb13++; nb23++;}
    if(nrarr[x][y-1] > nr) {nb14++; nb24++;}
    if((nrarr[x+2][y+1] > nr)) nb13++;
    if((nrarr[x+1][y+2] > nr)) nb13++;
    if((nrarr[x-2][y+1] > nr)) nb23++;
    if((nrarr[x-1][y+2] > nr)) nb23++;
    if((nrarr[x-2][y-1] > nr)) nb24++;
    if((nrarr[x-1][y-2] > nr)) nb24++;
    if((nrarr[x+2][y-1] > nr)) nb14++;
    if((nrarr[x+1][y-2] > nr)) nb14++;
    //band till tidigare
    if(lpnr(nrarr[x-1][y], nr)) {nb13--; nb14--;}
    if(lpnr(nrarr[x+1][y], nr)) {nb23--; nb24--;}
    if(lpnr(nrarr[x][y-1], nr)) {nb13--; nb23--;}
    if(lpnr(nrarr[x][y+1], nr)) {nb14--; nb24--;}
    if(lpnr(nrarr[x-2][y-1], nr)) nb13--;
    if(lpnr(nrarr[x-1][y-2], nr)) nb13--;
    if(lpnr(nrarr[x+2][y-1], nr)) nb23--;
    if(lpnr(nrarr[x+1][y-2], nr)) nb23--;
    if(lpnr(nrarr[x+2][y+1], nr)) nb24--;
    if(lpnr(nrarr[x+1][y+2], nr)) nb24--;
    if(lpnr(nrarr[x-2][y+1], nr)) nb14--;
    if(lpnr(nrarr[x-1][y+2], nr)) nb14--;
    if(nb13<0 || nb23<0 || nb24<0 || nb14<0) {std::cout << "Bug 3 in sfpossmin\n"; exit(0);}
	
    if(!nbrb) {
      switch(arr[x][y]) {
	//case 1: if((nd13 == 0 && nb13 > 1) || (nd14 == 0 && nb14 > 1)) return 0; break;
	//case 2: if((nd23 == 0 && nb23 > 1) || (nd24 == 0 && nb24 > 1)) return 0; break;
	case 1: if((nd13 == 0 && nb13 >= 1) || (nd14 == 0 && nb14 >= 1)){
	    deg = 1;
	    if((nd13 == 0 && nb13 > 1) || (nd14 == 0 && nb14 > 1)) return 0;
	  } break;
	case 2: if((nd23 == 0 && nb23 >= 1) || (nd24 == 0 && nb24 >= 1)) {
	    deg = 1;
	    if ((nd23 == 0 && nb23 > 1) || (nd24 == 0 && nb24 > 1)) return 0;
	  } break;
	case 3: if((nd13 == 0 && nb13 >= 1) || (nd23 == 0 && nb23 >= 1)) return 0; break;
	case 4: if((nd14 == 0 && nb14 >= 1) || (nd24 == 0 && nb24 >= 1)) return 0; break;
      }
    }
  }
  return 1;
}

int dfpossmin(int& deg)
{
  int x1,y1,x2,y2,x3,y3,x2p,y2p;
  x1 = y1 = START;
  x2 = y2 = START;
  movepos(x2,y2);
  x3 = x2;
  y3 = y2;
  movepos(x3,y3);
  deg = 0;

  do{
    x2p = x1 + x3 - x2;
    if(x2p != x2 && nneib(x2,y2) == 2){
      y2p = y1 + y3 - y2;
      if(!arr[x2p][y2p]){
	deg = 1;
	if(y2 > y2p || nneib(x2p,y2p) > 2) return 0;
      }
    }

    x1 = x2; y1 = y2;
    x2 = x3; y2 = y3;
  } while(movepos(x3,y3));

  return 1;
}

int nbond()
{
  int x = START, y = START;
  int n=0, pn, npnt=0;
  do {  //Every possible interaction is counted twice
    pn = n;
    if(arr[x+1][y] && arr[x][y]!=1 && arr[x+1][y]!=2) n++;
    if(arr[x-1][y] && arr[x][y]!=2 && arr[x-1][y]!=1) n++;
    if(arr[x][y+1] && arr[x][y]!=3 && arr[x][y+1]!=4) n++;
    if(arr[x][y-1] && arr[x][y]!=4 && arr[x][y-1]!=3) n++;
    if(pn != n) npnt++;
  } while(movepos(x,y));
  if(n & 1) std::cout << "Bug in nbond!\n";
  n >>= 1;

  fnbpont[npnt]++;
  if(npnt>nbpmax) nbpmax = npnt;
  return n;
}

void writeseq(std::ofstream &outf, int deg)
{
  static uint16_t seq[32], pkseq[4];
  int x=START, y=START, i;

  movepos(x,y);
  for(i=0; arr[x][y]<=4; i++){
    seq[i] = arr[x][y]-1;
    movepos(x,y);
  }
  if(i != nnod-2){ std::cout << "Bug in writeseq\n"; exit(0); }

  for(i=0; i<4; i++)
    pkseq[i]=0;
  for(i=0; i<nnod-2; i++){
    pkseq[i>>3] |= seq[i] << ((i & 7) << 1);
  }
  if(deg) pkseq[3] |= 0x8000;

  outf.write((char *)pkseq,8);
}

int selfavrek(int x, int y, int n, std::ofstream &outf)
{
  if(n == 2) {
    ldnposx = x;
    ldnposy = y;
  }

  int num = 0;
  if(n>1){
    nrarr[x][y] = nnod + 1 - n;
    arr[x][y] = 1;
    if(!arr[x+1][y]){
      arr[x][y] = 1;
      x++;
      num += selfavrek(x,y,n-1,outf);
      x--;
    }
    if(!arr[x-1][y]){
      arr[x][y] = 2;
      x--;
      num += selfavrek(x,y,n-1,outf);
      x++;
    }
    if(!arr[x][y+1]){
      arr[x][y] = 3;
      y++;
      upp++;
      num += selfavrek(x,y,n-1,outf);
      y--;
      upp--;
    }
    if(!arr[x][y-1] && upp){
      arr[x][y] = 4;
      y--;
      num += selfavrek(x,y,n-1,outf);
      y++;
    }
    arr[x][y] = 0;
    nrarr[x][y] = 0;
    return num;
  }
  else {
    nrarr[x][y] = nnod;
    arr[x][y] = 5;
    ntot++;

    int dfdeg, sfdeg;
    int pm = !dfrees() && !dfreee() && dfpossmin(dfdeg) && sfpossmin(sfdeg);
    if(pm) {
      int nb = nbond();
      fnbond[nb]++;
      if(nb>nbmax) nbmax = nb;
      if(!dfdeg && !sfdeg) nnotdeg++;

      writeseq(outf, dfdeg || sfdeg);
    }


    arr[x][y] = 0;
    nrarr[x][y] = 0;
    return pm;
  }
}

int main(int argc, char **argv)
{
   int x,y;
   char outfstr[16];

   ntot = 0;
   upp = 0;
   nnotdeg = 0;

	if(argc != 2)
	{
		std::cerr << "Usage: "<<argv[0]<<" <N>\n";
		exit(1);
	}
	nnod = atoi(argv[1]);

   strcpy(outfstr, "sa9n00.out");
   outfstr[4] = '0' + nnod / 10;
   outfstr[5] = '0' + nnod % 10;
   std::cout << "Output saved in " << outfstr << "\n";
   std::ofstream outf(outfstr, std::ios::out|std::ios::binary|std::ios::trunc);
   if(!outf){std::cerr << "Cannot create " << outfstr; exit(0);}

   for(x=0; x<ARRW; x++)
     for(y=0; y<ARRW; y++)
        arr[x][y]=0;

   for(x=0; x<ARRW; x++){
     fnbond[x]=0;
     fnbpont[x]=0;
   }
   nbmax=0;
   nbpmax=0;

   x = y = START;
   arr[x][y] = 1;
   nrarr[x][y] = 1;
   x++;

   std::cout << selfavrek(x,y,nnod-1,outf) << " possible minimal walks, ";
   std::cout << nnotdeg << " non-degenerate out of " << ntot << " walks\n";

   std::cout << "Frequency of binding sites\n";
   for(x=0; x<=nbmax; x++){
     std::cout << x << ": " << fnbond[x] << "\n";
   }
   int bsum=0;
   for(x=0; x<=nbmax; x++){
     bsum += fnbond[x]*x;
   }
   std::cout << "Total binding sites: " << bsum << "\n";

   outf.close();
}

