#ifndef LOGSTUFF_H
#define LOGSTUFF_H

#ifndef LOGFUNC
	#error You must define LOGFUNC to be the name of a logarithm function
#endif

#include <cmath>
#include <algorithm>

#define M_HALFLN2PI 0.918938533204672669540968854562379419803619384765625l

template<class prec, int max_n>
class logfactorial
{
  protected:
	prec factable[max_n + 1];

	prec logfac_calc(int n)
	{
		prec pn(n);
		return M_HALFLN2PI + (pn + .5) * LOGFUNC(pn) - pn + 1./(12.*pn);
	}
  public:
	logfactorial()
	{
		const int exacts = std::min(64, max_n + 1);
		prec f = 1;
		factable[0] = 0;
		for(int i = 1; i < exacts; i++)
		{
			f *= i;
			factable[i] = LOGFUNC(f);
		}
		for(int i = exacts; i <= max_n; i++)
			factable[i] = logfac_calc(i);
	}

	prec operator()(int n)
	{
		if(n > max_n)
			return logfac_calc(n);
		return factable[n];
	}

	prec ln_binom(int n, int k)
	{
	  return (*this)(n) - (*this)(k) - (*this)(n-k);
	}
};

template<class prec>
prec ln_add(prec a, prec b)
{
  if(!finite(a)) return b;
  if(!finite(b)) return a;
  if(a>b)
    return a + LOGFUNC(1. + exp(b-a));
  else
    return b + LOGFUNC(1. + exp(a-b));
}

#endif

