/****************************************************************************/
/* rmsd calculation                                                         */
/* flag = 0, backbone atoms (N,Ca,C)                                        */
/* flag = 1, all atoms (N,Ca,C,O + heavy sidechain atoms)                   */
/****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#define NMAX 150 
#define NTMAX 1000
#define NBMAX 3*NMAX
#define S1 " N  "," CA "," C  "," O  "
#define S2 " N  "," CA "," C  "," O  "," CB "
#define E "*"

const char atom[20][15][5]={
{S1,E},                                                               /* gly */
{S2,E},                                                               /* ala */
{S2," CG1"," CG2",E},                                                 /* val */
{S2," CG "," CD1"," CD2",E},                                          /* leu */
{S2," CG1"," CG2"," CD1",E},                                          /* ile */
{S2," OG ",E},                                                        /* ser */
{S2," OG1"," CG2",E},                                                 /* thr */
{S2," SG ",E},                                                        /* cys */
{S2," CG "," SD "," CE ",E},                                          /* met */
{S2," CG "," CD ",E},                                                 /* pro */
{S2," CG "," OD1"," OD2",E},                                          /* asp */
{S2," CG "," OD1"," ND2",E},                                          /* asn */
{S2," CG "," CD "," OE1"," OE2",E},                                   /* glu */
{S2," CG "," CD "," OE1"," NE2",E},                                   /* gln */
{S2," CG "," CD "," CE "," NZ ",E},                                   /* lys */
{S2," CG "," CD "," NE "," CZ "," NH1"," NH2",E},                     /* arg */
{S2," CG "," ND1"," CD2"," CE1"," NE2",E},                            /* his */
{S2," CG "," CD1"," CD2"," CE1"," CE2"," CZ ",E},                     /* phe */
{S2," CG "," CD1"," CD2"," CE1"," CE2"," CZ "," OH ",E},              /* tyr */
{S2," CG "," CD1"," CD2"," NE1"," CE2"," CE3"," CZ2"," CZ3"," CH2",E} /* trp */
};

const char aminoacid[20][4]={"GLY","ALA","VAL","LEU","ILE","SER","THR",
			     "CYS","MET","PRO","ASP","ASN","GLU","GLN",
			     "LYS","ARG","HIS","PHE","TYR","TRP"};

static double xnat[NTMAX],ynat[NTMAX],znat[NTMAX],gyr2nat;
static double xnatb[NBMAX],ynatb[NBMAX],znatb[NBMAX],gyr2natb;
static int nall=0,nba=0;
static int nha[NMAX];         /* # of heavy atoms */ 

void read_pdb(char *fn,int *nha,int nr1,int nr2,
	      double x[],double y[],double z[]);
double center_of_mass(double x[],double y[],double z[],int natom);
double correlation(double x1[],double y1[],double z1[],
		   double x2[],double y2[],double z2[],int natom);
int fgetline(char *line,int max,FILE *fp);
void substr(char *sub,char *str,int pos, int offset);

/****************************************************************************/
double rmsd2(int flag,double x[],double y[], double z[],int natom){
  double gyr2;
  if(flag==1){
   if(natom!=nall){printf("natom wrong -- natom=%i %i\n",natom,nall);exit(-1);}

    gyr2=center_of_mass(x,y,z,nall);
    return gyr2nat+gyr2-correlation(xnat,ynat,znat,x,y,z,nall);
  }
  if(natom!=nba){printf("natom wrong -- natom=%i\n",natom);exit(-1);}

  gyr2=center_of_mass(x,y,z,nba);
  return gyr2natb+gyr2-correlation(xnatb,ynatb,znatb,x,y,z,nba);
}
/****************************************************************************/
double rmsd2_2(double x1[],double y1[], double z1[],
	       double x2[],double y2[], double z2[],int n){
  double gyr2a,gyr2b;
  gyr2a=center_of_mass(x1,y1,z1,n);
  gyr2b=center_of_mass(x2,y2,z2,n);
  return gyr2a+gyr2b-correlation(x1,y1,z1,x2,y2,z2,n);
}
/****************************************************************************/
void rmsd_init(char *fn,int *seq,int nr1,int nr2){
  int i=0,j,aj=0,na=0;

  for(j=0;j<=nr2;j++){
    switch (seq[j]){  
      case 'G' : {nha[j]=4; break;}
      case 'A' : {nha[j]=5; break;}
      case 'V' : {nha[j]=7; break;}
      case 'L' : {nha[j]=8; break;}
      case 'I' : {nha[j]=8; break;}
      case 'S' : {nha[j]=6; break;}
      case 'T' : {nha[j]=7; break;}
      case 'C' : {nha[j]=6; break;}
      case 'M' : {nha[j]=8; break;}
      case 'P' : {nha[j]=7; break;}
      case 'D' : {nha[j]=8; break;}
      case 'N' : {nha[j]=8; break;}
      case 'E' : {nha[j]=9; break;}
      case 'Q' : {nha[j]=9; break;}
      case 'K' : {nha[j]=9; break;}
      case 'R' : {nha[j]=11; break;}
      case 'H' : {nha[j]=10; break;}
      case 'F' : {nha[j]=11; break;}
      case 'Y' : {nha[j]=12; break;}
      case 'W' : {nha[j]=14; break;}
      default : {printf("%i -- nha\n",j);exit(-1);}
    }
  }   

  read_pdb(fn,nha,nr1,nr2,xnat,ynat,znat);
  for(j=nr1;j<=nr2;j++){
    xnatb[i]=xnat[aj]; ynatb[i]=ynat[aj]; znatb[i]=znat[aj]; i++; aj++;
    xnatb[i]=xnat[aj]; ynatb[i]=ynat[aj]; znatb[i]=znat[aj]; i++; aj++;
    xnatb[i]=xnat[aj]; ynatb[i]=ynat[aj]; znatb[i]=znat[aj]; i++; aj++;
    aj+=nha[j]-3;
    na+=nha[j];
  }
  nba=i;
  if(nall!=na){printf("rmsd_init --  nall=%i\n",nall); exit(-1);}
  if(nba>NBMAX){printf("nba too big -- nba=%i\n",nba);exit(-1);}      
  if(nall>NTMAX){printf("nall too big -- nall=%i\n",nall);exit(-1);}      

  gyr2nat=center_of_mass(xnat,ynat,znat,nall);
  gyr2natb=center_of_mass(xnatb,ynatb,znatb,nba);
  printf("gyr2 for target: (all) %f (backbone) %f \n",gyr2nat,gyr2natb);
}
/****************************************************************************/
void read_pdb(char *fn,int *nha,int nr1,int nr2,
	      double x[],double y[],double z[])
{
  FILE *fp;
  int i=0,j=0,iaa=0,sq=-1;
  char line[100],spec[8],atomname[5],amino[4],coord[9];
  fp=fopen(fn,"r");

  fgetline(line,100,fp);
  substr(spec,line,0,6);
  substr(atomname,line,12,4);
  substr(amino,line,17,3);
  while(strcmp(spec,"ATOM  ")!=0){
    fgetline(line,100,fp);
    substr(spec,line,0,6);
    substr(atomname,line,12,4);
    substr(amino,line,17,3);
  }

  while(strcmp(spec,"ATOM  ")==0){
    if(strcmp(atomname," N  ")==0){
      if((sq>-1)&&(j!=nha[sq]))
	{printf("in delta2.c, seq no. %i\n",sq); exit(-1);}
      if(sq==nr2) break;
      j=0;
      sq++;
      iaa=0;
      while(strcmp(aminoacid[iaa],amino)!=0)
	if(++iaa>19){printf("error -- unknown amino acid\n"); exit(-1);}
    }
    if(strcmp(atomname,atom[iaa][j])==0){
      if(sq>=nr1){
	substr(coord,line,31,8); x[i]=atof(coord);
	substr(coord,line,39,8); y[i]=atof(coord);
	substr(coord,line,47,8); z[i]=atof(coord);
	i++;
      }
      j++;
    }
    if(fgetline(line,100,fp)==0) break;
    substr(spec,line,0,6);
    substr(atomname,line,12,4);      
    substr(amino,line,17,3);
  }
  nall=i;
  if(j!=nha[sq]){printf("in delta2.c, seq no. %i\n",sq); exit(-1);}
  fclose(fp);
}
/****************************************************************************/
double center_of_mass(double x[],double y[],double z[], int natom)
{
  double xcm,ycm,zcm,gyr2;
  int j;
  xcm=ycm=zcm=0;
  for (j=0;j<natom;j++) {xcm+=x[j]; ycm+=y[j]; zcm+=z[j];}
  xcm*=1.0/natom; ycm*=1.0/natom; zcm*=1.0/natom;
  gyr2=0;
  for (j=0;j<natom;j++){
    x[j]-=xcm; y[j]-=ycm; z[j]-=zcm;
    gyr2+=x[j]*x[j]+y[j]*y[j]+z[j]*z[j];
  }
  return gyr2/natom;
}
/****************************************************************************/
double correlation(double x1[],double y1[],double z1[],
		   double x2[],double y2[],double z2[],int natom)
{
  double R[3][3],detR;
  double RtR[3][3],a,b,c,q,r,theta,w[3];
  double pi2=2*acos(-1.);
  int i,j,tmp;
 
  for (i=0;i<3;i++) {for (j=0;j<3;j++) R[i][j]=0;}
  for (i=0;i<natom;i++) {
    R[0][0]+=x1[i]*x2[i];
    R[0][1]+=x1[i]*y2[i];
    R[0][2]+=x1[i]*z2[i];
    R[1][0]+=y1[i]*x2[i];
    R[1][1]+=y1[i]*y2[i];
    R[1][2]+=y1[i]*z2[i];
    R[2][0]+=z1[i]*x2[i];
    R[2][1]+=z1[i]*y2[i];
    R[2][2]+=z1[i]*z2[i];
  }
  detR=+R[0][0]*R[1][1]*R[2][2]
       +R[0][1]*R[1][2]*R[2][0]
       +R[0][2]*R[1][0]*R[2][1]
       -R[2][0]*R[1][1]*R[0][2]
       -R[2][1]*R[1][2]*R[0][0]
       -R[2][2]*R[1][0]*R[0][1];
  if(detR==0){printf("****** WARNING! *******\ndetR=0\n");}
  

  for (i=0;i<3;i++) {
    for (j=0;j<3;j++) {
       RtR[i][j]=R[0][i]*R[0][j]+R[1][i]*R[1][j]+R[2][i]*R[2][j];
    }
  }

  a=-(RtR[0][0]+RtR[1][1]+RtR[2][2]);
  b=RtR[0][0]*RtR[1][1]+RtR[0][0]*RtR[2][2]+RtR[1][1]*RtR[2][2]-
    RtR[0][1]*RtR[0][1]-RtR[0][2]*RtR[0][2]-RtR[1][2]*RtR[1][2];
  c=-RtR[0][0]*RtR[1][1]*RtR[2][2]
    -RtR[0][1]*RtR[1][2]*RtR[2][0]
    -RtR[0][2]*RtR[1][0]*RtR[2][1]
    +RtR[2][0]*RtR[1][1]*RtR[0][2]
    +RtR[2][1]*RtR[1][2]*RtR[0][0]
    +RtR[2][2]*RtR[1][0]*RtR[0][1];
/*  c=-detR*detR;*/

  q=(a*a-3*b)/9;
  r=(2*a*a*a-9*a*b+27*c)/54;
  if (r*r/(q*q*q)>=1) {
    printf("error r^2=%e q^3=%e %e\n",r*r,q*q*q,r*r/(q*q*q));
    exit(-1);
  }
  q=sqrt(q);
  theta=acos(r/(q*q*q));
  w[0]=sqrt(fabs(-2*q*cos(theta/3)-a/3));
  w[1]=sqrt(fabs(-2*q*cos((theta+pi2)/3)-a/3));
  w[2]=sqrt(fabs(-2*q*cos((theta-pi2)/3)-a/3));
  
  if(detR<0){
    tmp = w[1] > w[0] ? 1 : 0;
    return( 2.0/natom*(w[tmp]+fabs(w[(tmp+1)%3]-w[(tmp+2)%3])) );
  }else
    return( 2.0/natom*(w[0]+w[1]+w[2]) );
}
/****************************************************************************/
void substr(char *sub,char *str,int pos, int offset)
{
  strncpy(sub,str+pos,offset);
  *(sub+offset)='\0';
}
/****************************************************************************/
int fgetline(char *line,int max,FILE *fp)
{
  if(fgets(line,max,fp) == NULL){
     strcpy(line,"");
     return 0;
  }
  else
     return strlen(line);
}
/****************************************************************************/

#undef NMAX
#undef NTMAX
#undef NBMAX
#undef S1 
#undef S2
#undef E











