#!/usr/local/bin/php
<?


function html($s)
{
	if($s == "")
		return "&nbsp;";
	return htmlspecialchars($s);
}

function coloredBackground(&$arr)
{
	if($arr[4] == -1 || $arr[3] <= 0)
		return "bgcolor='#d0d0ff'";
	if($arr[4] / $arr[3] > .025)
		return "bgcolor='#ffb0b0'";
	if($arr[4] / $arr[3] > .010)
		return "bgcolor='#ffd0d0'";
	return "";
}

function coloredNumber($val, $exp)
{
	$ival = (int)($val + .5);
	if($exp < 0)
		return $ival;
	$f = log($val / $exp);
	if($f > 0.3)
		$c = "#00a000";
	else if($f > 0.15)
		$c = "#005000";
	else if($f < -0.3)
		$c = "#ff0000";
	else if($f < -0.15)
		$c = "#a00000";
	else
		return $ival;
	return "<font color='$c'>$ival</font>";
}

function computerSorter($c1, $c2)
{
	if($c1[3] > $c2[3])
		return -1;
	if($c1[3] < $c2[3])
		return 1;
	if($c1[0] > $c2[0])
		return -1;
	if($c1[0] < $c2[0])
		return 1;
	return 0;
}

function maketable($fd)
{
	$tests = array(); // test => computers
	$comps = array();	// computer => array(MHz, cpu, def-flags, total_score,
							// compiler, parallelism)
	$times = array();	// flags => array(computer =>
							// array(test => array(time, n, time^2, mean, sd, min)))

	$ignoreflags = array_flip(array("-march=[best match]",
		"-march=i686",
		"-march=i686 -fomit-frame-pointer -maccumulate-outgoing-args -mtune=pentium4",
		"-march=[best match] -fomit-frame-pointer -maccumulate-outgoing-args",
		"-march=[best match] -fomit-frame-pointer -maccumulate-outgoing-args -ffast-math"));
	$mapflags = array(
		"-march=i686" =>
		"",
		"-fomit-frame-pointer -maccumulate-outgoing-args" =>
		"plus -fomit-frame-pointer -maccumulate-outgoing-args",
		"-march=i686 -fomit-frame-pointer -maccumulate-outgoing-args" =>
		"plus -fomit-frame-pointer -maccumulate-outgoing-args",
		"-fomit-frame-pointer -maccumulate-outgoing-args -ffast-math" =>
		"plus -ffast-math",
		"-march=i686 -fomit-frame-pointer -maccumulate-outgoing-args -ffast-math" =>
		"plus -ffast-math",
		"profiling -fomit-frame-pointer -maccumulate-outgoing-args -ffast-math" =>
		"plus profiling for branch prediction",
		"profiling -march=i686 -fomit-frame-pointer -maccumulate-outgoing-args -ffast-math" =>
		"plus profiling for branch prediction",
		"profiling -march=[b.m.] -fomit-frame-pointer -maccumulate-outgoing-args -ffast-math" =>
		"plus -march=[best match]"
	);

	$weights = array("4cyc" => .5, "sand" => .3, "6cyc" => .5, "6c_ld" => .5);

	foreach($mapflags as $f)
		$times[$f] = array();

	$curcomp = "";
	for($line = 1; ($row = fgets($fd, 1024)) !== false; $line++)
	{
		$row = rtrim($row);
		if(ereg("^([^:]+) : ([^:]+) : ([0-9]+) MHz (.*) : (.*)\$",
			$row, $regs))
		{
			if($curcomp != "" && $curflags == "" && !$times[""][$curcomp])
			{
				unset($times[""][$curcomp]);
				unset($comps[$curcomp]);
			}

			$curcomp = $regs[1];
			$compiler = $regs[2];
			$curflags = "";
			$mhz = $regs[3];
			$cpuarch = $regs[4];
			$gccflags = $regs[5];
			$parallel = 1.;
			if(ereg("\\*([0-9]+)/", $curcomp, $regs))
				$parallel = (float)$regs[1];

			if(isset($comps[$curcomp]))
			{
				for($i = 1; isset($comps[$curcomp.";$i"]); $i++) {}
				$curcomp .= ";$i";
			}
			$comps[$curcomp] = array($mhz, $cpuarch, $gccflags, 0, $compiler, $parallel);
			$times[""][$curcomp] = array();
		}
		else if($curcomp == "")
		{
			fwrite(STDERR, "Test time without computer on line $line\n");
			return false;
		}
		else if(ereg("^([^: ]+): ([0-9.]+)\$", $row, $regs))
		{	
			$test = $regs[1];
			$time = (float)$regs[2];
			$time2 = $time * $time;
			if($time <= 0)
			{
				fwrite(STDERR, "Non-positive time on line $line\n");
				return false;
			}
			if(!isset($tests[$test]))
			{
				if($curflags == "")
					$tests[$test] = 0;
				else
				{
					fwrite(STDERR,
						"Unknown test for non-default flags on line $line\n");
					return false;
				}
			}
			$arr =& $times[$curflags][$curcomp];
			if(!isset($arr[$test]))
			{
				$arr[$test] = array(0.0, 0, 0.0, 0.0, 0.0, $time);
				if($curflags == "")
					$tests[$test]++;
			}
			else
				$arr[$test][5] = min($arr[$test][5], $time);
			$w = isset($weights[$test]) ? $weights[$test] : 1.;
			$arr[$test][0] += $time;
			$arr[$test][1]++;
			$arr[$test][2] += $time2;
			unset($arr);
		}
		else if(ereg("^New compiler flags: (.*)", $row, $regs))
		{
			if($regs[1] == "(none)")
				$curflags = "Default target";
			else if(isset($mapflags[$regs[1]]))
				$curflags = $mapflags[$regs[1]];
			else
				$curflags = "$regs[1]";

			if(!isset($times[$curflags]))
				$times[$curflags] = array();
			if(!isset($times[$curflags][$curcomp]) || $curflags == "")
				$times[$curflags][$curcomp] = array();
		}
		else if(ereg("^Output matches", $row))
		{
		}
		else if(ereg("^Output differs!", $row))
		{
			fwrite(STDERR, "WARNING ($curcomp): \"$row\"\n");
		}
		else
		{
			fwrite(STDERR, "Malformatted input on line $line: \"$row\"\n");
//			return false;
		}
	}
	if($curcomp != "" && $curflags == "" && !$times[""][$curcomp])
	{
		unset($times[""][$curcomp]);
		unset($comps[$curcomp]);
	}

	if(!$tests)
	{
		fwrite(STDERR, "No data in input file\n");
		return false;
	}

	unset($arr, $arr2);
	for(reset($times); list($flag) = each($times); )
	{
		for(reset($times[$flag]); list($comp) = each($times[$flag]); )
		{
			$arr =& $times[$flag][$comp];
			for(reset($arr); list($test) = each($arr); )
			{
				$arr2 =& $arr[$test];
				$mean = $arr2[0] / $arr2[1];
				$arr2[3] = $mean;
//				$arr2[3] = $arr2[5];		// minimum time
				if($arr2[1] > 1)
				{
					$sd = sqrt(max(0, $arr2[2] / $arr2[1] - $mean * $mean)) /
						($arr2[1] - 1);
				}
				else
					$sd = -1;
				$arr2[4] = $sd;
			}
		}
	}

	$normtimes = array();	// comp => array(test => factor)
	$normmhz = array();	// test => mhz
	$testcnt = count($tests);
	for(reset($tests); list($test, $timecnt) = each($tests); )
	{
		$timesum = 0;
		$mhzsum = 0;
		$cnt = 0;
		for(reset($comps); list($comp) = each($comps); )
		{
			if(count($times[""][$comp]) == $testcnt)
			{
				$f = ($comps[$comp][0] == 800 &&
					substr($comps[$comp][4], 0, 3) == "4.1") ? 1. : .0001;
				$timesum += log($times[""][$comp][$test][3]) * $f;
				$mhzsum += log($comps[$comp][0]) * $f;
				$cnt += $f;
			}
		}
		// Geometric mean of execution times.
		$timeavg = exp($timesum / $cnt);
		$normmhz[$test] = exp($mhzsum / $cnt);
		for(reset($comps); list($comp) = each($comps); )
		{
			if(!isset($normtimes[$comp]))
				$normtimes[$comp] = array();
			if(isset($times[""][$comp][$test]))
			{
				$normtimes[$comp][$test] =
					$times[""][$comp][$test][3] / $timeavg;
			}
		}
	}

	$bogomhz = array();	// flags => array(comp => bogomhz)
	$totweight = 0.;
	foreach($tests as $test => $dummy)
		$totweight += (isset($weights[$test]) ? $weights[$test] : 1.);
	foreach($times as $flag => $ftimes)
	{
		$bogomhz[$flag] = array();
		foreach($ftimes as $comp => $ttimes)
		{	
			// Only calculate if all tests have been run
			if(count($ttimes) != $testcnt)
				continue;
			$timesum = 0;
			foreach($ttimes as $test => $time)
			{
				$w = isset($weights[$test]) ? $weights[$test] : 1.;
				$bogo = $normmhz[$test] / $normtimes[$comp][$test] *
					$times[""][$comp][$test][3] / $time[3];
				$timesum += log($bogo) * $w;
			}
			$parallel = $comps[$comp][5];
			$bogomhz[$flag][$comp] = $parallel * exp($timesum / $totweight);
		}
	}

	for(reset($comps); list($comp) = each($comps); )
	{
		$comps[$comp][3] =
			isset($bogomhz[""][$comp]) ? $bogomhz[""][$comp] : -1;
	}

	uasort($comps, "computerSorter");

	$testheads = "<th nowrap>".implode("</th>\n<th nowrap>",
		array_keys($tests))."</th>";

	$flagheads1 = "";
	$flagheads2 = "";
	$colsperflag = count($tests) + 2;
	foreach($times as $flag => $dummy)
	{
		if(isset($ignoreflags[$flag]))
			continue;
		if($flag != "")
		{
			$flagheads1 .= "<th nowrap colspan=$colsperflag>$flag</th>";
			$flagheads2 .= "<th nowrap>ALL</th><th nowrap>% ch</th>$testheads";
		}
	}

?>
<table border=1 cellpadding=2>
<tr><th rowspan=2>Computer</th><th rowspan=2>CPU</th>
	<th rowspan=2>MHz</th>
	<th rowspan=2>GCC</th>
	<th rowspan=2>-march</th>
	<th nowrap colspan=<?= $testcnt+2 ?>>Score (bogo-MHz) (with -march=i686)</th>
	<?= $flagheads1 ?></tr>
<tr><th>ALL</th><th>/MHz</th>
	<?= $testheads ?> 
	<?= $flagheads2 ?></tr>
<?

	$tdbg = array("bgcolor='#c0c0c0'", "bgcolor='#e0e0e0'");
	for(reset($comps), $odd = 0; list($comp, $cinfo) = each($comps); $odd ^= 1)
	{
		list($mhz, $cpuarch, $gccflags, , $compiler, $parallel) = $cinfo;
		$march = ereg_replace(".*-march=([^ ]*).*", "\\1", $gccflags);
		echo "<tr $tdbg[$odd]><th nowrap>".html($comp)."</th>\n".
			"<td nowrap>".html($cpuarch)."</td>\n".
			"<td>".html($mhz)."</td>\n".
			"<td>".html($compiler)."</td>\n".
			"<td nowrap>".html($march)."</td>\n";
		$tm =& $times[""][$comp];
		$nt =& $normtimes[$comp];

		for(reset($times); list($flag) = each($times); )
		{
			if(isset($ignoreflags[$flag]))
				continue;
			if(isset($bogomhz[$flag][$comp]))
			{
				$bogo = (int)($bogomhz[$flag][$comp] + .5);
				if($flag == "")
					$bogox = sprintf("%.2f", $bogomhz[""][$comp] / $mhz);
				else
				{
					$brel = $bogomhz[$flag][$comp] / $bogomhz[""][$comp] - 1.;
					$bogox = sprintf("%+.1f%%", 100 * $brel);
				}
			}
			else
				$bogo = $bogox = "-";
			echo "<td align=right><b>$bogo</b></td>\n";
			echo "<td nowrap align=right>$bogox</td>\n";
			for(reset($tests); list($test) = each($tests); )
			{
				if(!isset($nt[$test]) || !isset($times[$flag][$comp]) ||
					!isset($times[$flag][$comp][$test]) ||
					$times[$flag][$comp][$test][3] <= 0)
				{
					echo "<td align=right>-</td>\n";
				}
				else
				{
					$bg = coloredBackground($times[$flag][$comp][$test]);
					$bogo = $parallel * $normmhz[$test] / $normtimes[$comp][$test] *
						$times[""][$comp][$test][3] / $times[$flag][$comp][$test][3];
					if(isset($bogomhz[$flag][$comp]))
						$cn = coloredNumber($bogo, $bogomhz[$flag][$comp]);
					else
						$cn = coloredNumber($bogo, -1);
					echo "<td align=right $bg>$cn</td>\n";
				}
			}
		}

		echo "</tr>\n";
	}
	echo "</table>\n";
}

$fd = null;
if(isset($argv[1]))
{
	$fd = fopen($argv[1], "r") or die("failed to open $argv[1]\n");
	while(($r = fgets($fd)) !== false)
	{
		echo $r;
		if(strpos($r, "BEGIN") !== false) break;
	}
}
$ok = maketable(STDIN);
if($fd)
{
	$skip = true;
	while(($r = fgets($fd)) !== false)
	{
		if(strpos($r, "END") !== false)
			$skip = false;
		if(!$skip)
			echo $r;
	}
}

return !$ok;

?>
