/**
 * Filename     : baseCompartmentChange.cc
 * Description  : A base class describing variable updates
 * Author(s)    : Henrik Jonsson (henrik@thep.lu.se)
 * Created      : October 2003
 * Revision     : $Id: baseCompartmentChange.cc,v 1.12 2006/03/18 00:05:14 henrik Exp $
 */
#include<vector>

#include"baseCompartmentChange.h"
//#include"compartmentDivision.h"
//#include"compartmentRemoval.h"

BaseCompartmentChange::~BaseCompartmentChange() {
  std::cerr << "BaseCompartmentChange::~BaseCompartmentChange()\n";
}

//!Factory creator, all creation should be mapped onto this one 
/*! Given the idValue a compartmentChange of the defined type is returned
(using new Class).*/
BaseCompartmentChange *
BaseCompartmentChange::
createCompartmentChange(std::vector<double> &paraValue,
			std::vector< std::vector<size_t> > &indValue, 
			std::string idValue ) {
  
  //Divisions
  //(compartmentDivision.h,compartmentDivision.cc)
  //////////////////////////////////////////////////////////////////////
  //Removals
  //(compartmentRemoval.h,compartmentRemoval.cc)
  //////////////////////////////////////////////////////////////////////
  //Default, if nothing found
  //////////////////////////////////////////////////////////////////////
  {
    std::cerr << "\nBaseCompartmentChange::createCompartmentChange() "
	      << "WARNING: CompartmentChangetype " 
	      << idValue << " not known, no compartmentChange created.\n\7";
    exit(-1);
  }
}

//!This creator reads from an open file and then calls for the main creator
BaseCompartmentChange * 
BaseCompartmentChange::createCompartmentChange(std::istream &IN ) {
  
  std::string idVal;
  size_t pNum,levelNum;
  IN >> idVal;
  IN >> pNum;
  IN >> levelNum;
  std::vector<size_t> varIndexNum( levelNum );
  for( size_t i=0 ; i<levelNum ; i++ )
    IN >> varIndexNum[i];

  std::vector<double> pVal( pNum );
  for( size_t i=0 ; i<pNum ; i++ )
    IN >> pVal[i];

  std::vector< std::vector<size_t> > varIndexVal( levelNum );
  for( size_t i=0 ; i<levelNum ; i++ )
    varIndexVal[i].resize( varIndexNum[i] );

  for( size_t i=0 ; i<levelNum ; i++ )
    for( size_t j=0 ; j<varIndexNum[i] ; j++ )
      IN >> varIndexVal[i][j];

  return createCompartmentChange(pVal,varIndexVal,idVal);
}

int BaseCompartmentChange::
flag(Compartment &compartment,std::vector< std::vector<double> > &y,
     std::vector< std::vector<double> > &dydt ) {
  std::cerr << "BaseCompartmentChange::flag() should not be used. "
	    << "Should always be mapped onto one of the real types.\n";
  exit(0);
}  

void BaseCompartmentChange::
update(Organism &O,Compartment &compartment,
       std::vector< std::vector<double> > &y,
       std::vector< std::vector<double> > &dydt, double t ) {
  std::cerr << "BaseCompartmentChange::update() should not be used. "
	    << "Should always be mapped onto one of the real types.\n";
  exit(0);
}  
