/**
 * Filename     : species.cc
 * Description  : A class describing a species (typically a molecule)
 * Author(s)    : Henrik Jonsson (henrik@thep.lu.se)
 * Created      : October 2003
 * Revision     : $Id: species.cc,v 1.4 2006/03/18 12:23:08 henrik Exp $
 */

#include<iostream>
#include<fstream>
#include "species.h"

Species::Species() {
  //std::cerr << "Species::Species() not defined.\n";
  setIndex(static_cast<size_t>(-1));
}

Species::Species( const Species & speciesCopy ) { 
  
  setId( speciesCopy.id() );
  setIndex( speciesCopy.index() );
  setReaction( speciesCopy.reaction() );
}

Species::Species(std::vector<BaseReaction*> &reactionValue, 
		 std::string idValue, size_t indexValue ) {
  std::cerr << "Species::Species(std::vector<BaseReaction*> "
	    << "&reactionValue,... not defined.\n";
}

Species::Species( char *inFile ) {
  
  std::ifstream IN( inFile );
  if( !IN ) {
    std::cerr << "Species::Species(char*) - "
	      << "Cannot open file " << inFile << "\n\n\7";exit(-1);}
  readSpecies(IN);
}

Species::Species( std::string inFile ) { 
  
  const char *tmp = inFile.c_str();
  std::ifstream IN( tmp );
  if( !IN ) {
    std::cerr << "Species::Species() - "
	      << "Cannot open file " << inFile << "\n\n\7";exit(-1);}
  readSpecies(IN);
}

//!Creates a species from reading in an open file (typically used by organism)
Species::Species( std::ifstream &IN ) {
  
  readSpecies(IN);
}

Species::~Species() {
  //std::cerr << "Species::~Species() not defined.\n";
}

//!Read a species from an open file 
void Species::readSpecies( std::ifstream &IN ) {
  
  //Remove all reactions before adding
  if( numReaction() ) 
    reaction_.resize(0);
  
  size_t indexVal,numReactionVal;
  std::string idVal;
  IN >> idVal;
  setId(idVal);
  IN >> indexVal;
  setIndex(indexVal);
  IN >> numReactionVal;
  for( size_t i=0 ; i<numReactionVal ; i++ ) {
    if( addReaction(IN) )
      std::cerr << "Species::readSpecies() Warning Adding reaction failed for "
		<< "species " << index() << " " << id() << "\n";
  }
}

//!Adds a reaction to the reaction vector from an open file
int Species::addReaction( std::ifstream &IN ) {
  if( !IN )
    return -1;
  reaction_.push_back( BaseReaction::createReaction(IN) );
  return 0;
}

//!Adds up all derivative contributions from the reactions
void Species::derivs(Compartment &compartment,
		     std::vector< std::vector<double> > &y,
		     std::vector< std::vector<double> > &dydt ) {
  
  for( size_t i=0 ; i<numReaction() ; i++ )
    reaction(i)->derivs(compartment,index(),y,dydt);
}
